/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

#include "eel.h"

/* insert the following extended ascii character into the buffer */

command quoted_insert() on reg_tab[CTRL('Q')]
{
	if (_pushed_special_key == -1) {
		mention("Quote: ");
		_now_quoting = 1;
		do getkey(); while (key >= 256);
		_now_quoting = 0;
		if (_pushed_special_key != -1)
			getkey();
		say("");
		if (user_abort == 2)
			check_abort();
		user_abort=0;
		while (iter-- > 0)
			insert(key);
	} else {
		getkey();
		user_abort=0;
		while (iter-- > 0)
			normal_character();
	}
}

/* toggle all self-inserting characters with self-overwrite */
command overwrite_mode() on reg_tab[KEYINSERT]
{
	over_mode = (has_arg? (iter != 0) : !over_mode);
	make_mode();
	iter = 1;
}

/* insert or overwrite according to mode */
command normal_character() on reg_tab[CTRL('J')]	/* more, see below */
{
	int i;

	check_abort();
	if (key >= 256) return;
	if (key == '\n') {
		if (over_mode) {
			if (!nl_forward())
				insert('\n');
		} else if (auto_indent && indenter) {
			delete(point, point - parse_string(-1, "[ \t]*"));
			insert('\n');		/* delete trailing whitesp */
			(*indenter)();
		} else
			insert('\n');
	} else if (over_mode && point < size() && curchar() != '\n') {
		if (curchar() == '\t')
			i = display_width('\t', current_column());
		else
			i = 0;
		replace(point++, key);
		while (--i > 0)		/* insert i-1 spaces after pt */
			insert(' '), point--;
	} else
		insert(key);
}

when_loading()		/* put it on the other keys, too */
{
	int i;

	for (i = '!'; i <= '~'; i++)
		reg_tab[i] = (short) normal_character;
}

/* Go to the beginning of the current line */
command beginning_of_line() on reg_tab[CTRL('A')], reg_tab[NUMALT(KEYLEFT)]
{
	iter=0;				/* throw away any argument */
	cursor_to_column = -1;
	to_begin_line();
}

/* Go to the end of the current line */
command end_of_line() on reg_tab[CTRL('E')], reg_tab[NUMALT(KEYRIGHT)]
{
	iter=0;			/* throw away any argument */
	cursor_to_column = -1;
	to_end_line();
}

give_begin_line()	/* return the position of this line's start */
{
	save_var point;
	to_begin_line();
	return point;
}

give_end_line()		/* return the position of this line's end */
{
	save_var point;
	to_end_line();
	return point;
}

command down_line() on reg_tab[CTRL('N')], reg_tab[KEYDOWN]
{
	int true;

	down_or_up_lines(1);
	true = point;
	while (scroll_at_end > 0 && true > window_end) {
		window_scroll(scroll_at_end);
		if (window_end + 1 >= size() - narrow_end)
			break;		/* make sure we stop at end */
		point = window_end + 1;	/* compute window_end explicitely */
		window_end = next_screen_line(scroll_at_end) - 1;
		check_abort();
	}
	point = true;
}

command up_line() on reg_tab[CTRL('P')], reg_tab[KEYUP]
{
	int max = iter + 5;

	down_or_up_lines(0);
	while (scroll_at_end > 0 && point < window_start) {
		window_scroll(-scroll_at_end);
		check_abort();
		if (max-- < 0)
			break;
	}
}

#if 0

virtual_column()
{
	return (cursor_to_column >= 0) ? cursor_to_column : current_column();
}

#else

virtual_column()
{
	return (virtual_space && cursor_to_column >= 0) ?
              cursor_to_column : current_column();
}

#endif

#if 1

same_column()
{
    if (goal_column >= 0)
        _preserve_column = goal_column;
    else if (virtual_space > 1 || prev_cmd != SAMECOL)
		_preserve_column = virtual_column();
}

keep_column()
{
    to_virtual_column(_preserve_column);
	this_cmd = SAMECOL;
}

#endif

#if 0

down_or_up_lines(down_lines)
{
	int hit_end = 0, wantcol;

	if (iter < 0)
		iter = -iter, down_lines = !down_lines;
	if (prev_cmd != SAMECOL)
		_preserve_column = virtual_space ? virtual_column()
					: current_column();
	while (iter-- > 0)
		if (!down_lines)
			nl_reverse();
		else if (!nl_forward()) {
			hit_end = 1;
			break;
		}
	wantcol = (goal_column >= 0) ? goal_column : _preserve_column;
	if (virtual_space)
		to_virtual_column(wantcol);
	else if (!hit_end || goal_column >= 0)	/* stay at end if we tried */
		move_to_column(wantcol);	/* to move past it */
	this_cmd = SAMECOL;
	iter = 0;
}

#else

down_or_up_lines(down_lines)
{
	if (iter < 0)
		iter = -iter, down_lines = !down_lines;
    same_column();
	while (iter-- > 0)
		if (!down_lines)
			nl_reverse();
		else if (!nl_forward()) {
            if (virtual_space == 0 && goal_column < 0)
                _preserve_column = current_column();
			break; /* stay at end if we tried to move past it */ 
		}
    keep_column();
	iter = 0;
}

#endif

#if 0

to_virtual_column(col)
{
	move_to_column(col);
	if (current_column() != col) {
		cursor_to_column = col;
		call_on_modify = 1;
	} else
		cursor_to_column = -1;
}   

#else

to_virtual_column(col)
{
	move_to_column(col);
    if (virtual_space == 0 || current_column() == col)
		cursor_to_column = -1;
    else {
		cursor_to_column = col;
		call_on_modify = 1;
	}
}

#endif 

#if 0

/* Set point to the previous character in the buffer */
command backward_character() on reg_tab[CTRL('B')], reg_tab[KEYLEFT]
{
	if (virtual_space < 2) {
		point -= iter;
		iter = 0;
	} else
		to_virtual_column(virtual_column() - 1);
}

#else

/* Set point to the previous character in the buffer */
command backward_character() on reg_tab[CTRL('B')], reg_tab[KEYLEFT]
{
	if (virtual_space < 2)
		point -= iter;
	else
		to_virtual_column(virtual_column() - iter);
    iter = 0;
}


#endif

#if 0

/* Set the point to the next character in the buffer */
command forward_character() on reg_tab[CTRL('F')], reg_tab[KEYRIGHT]
{
	if (virtual_space < 2) {
		point += iter;
		iter = 0;
	} else
		to_virtual_column(virtual_column() + 1);
}  

#else

/* Set the point to the next character in the buffer */
command forward_character() on reg_tab[CTRL('F')], reg_tab[KEYRIGHT]
{
	if (virtual_space < 2)
		point += iter;
	else
		to_virtual_column(virtual_column() + iter);
    iter = 0;
}

#endif

command open_line() on reg_tab[CTRL('O')]
{
	insert('\n');
	point--;
}

command goto_beginning() on reg_tab[ALT('<')], reg_tab[NUMCTRL(KEYHOME)]
{
	point = 0;
}

command goto_end() on reg_tab[ALT('>')], reg_tab[NUMCTRL(KEYEND)]
{
	point = size();
}

go_line(line)
{
	point = 0;
	while (--line > 0)
		if (!nl_forward())
			break;
}

command goto_line() on cx_tab['g']
{
	int num = get_number("Go to line: ");

	if (!got_bad_number)
		go_line(num);
}

full_getkey(prompt, code)	/* get a key using ^Q, ^[, and ^^ prefixes */
char *prompt;
{
	char msg[80];
	int allcode = 0;

	strcpy(msg, prompt);
	for (;;) {
		switch (code) {
			case ALTIFY_KEY:  strcat(msg, "<Alt> "); break;
			case CTRLIFY_KEY: strcat(msg, "<Ctrl> "); break;
		}
		allcode |= code, code = 0;
		for (;;) {
			mention("%s", msg);
			getkey();
			if (key != MOUSE_MOVE)
				break;
			maybe_refresh();
		}
		if (allcode & ALTIFY_KEY)
			key = make_alt(key);
		if (allcode & CTRLIFY_KEY)
			key = make_ctrl(key);
		if (root_keys[key] == (short) alt_prefix)
			code = ALTIFY_KEY;
		else if (root_keys[key] == (short) ctrl_prefix)
			code = CTRLIFY_KEY;
		else
			return;
	}
}

command insert_ascii() on reg_tab[ALT('#')] /* insert character by number */
{
	int res;

	if (run_by_mouse)
		has_arg = 0;
	res = get_number("Insert character with ASCII code: ");
	if (res < 0 || res > 255)
		error("Invalid ASCII code %d.", res);
	else if (!got_bad_number)
		insert(res);
}

short *ask_key(pr, nospace, keyname)	/* prompt for a key, giving name */
	char *pr, *keyname;		/* and pointer for binding */
{
	int index;
	short *tab, *alttab;
	char spare[80], *space, *spaceless, msg[80];

	spaceless = spare;
	space = keyname;
	if (nospace) {
		spaceless = keyname;
		space = spare;
	}
	tab = mode_keys;
	alttab = root_keys;
	*space = *spaceless = 0;
	for (;;) {
		sprintf(msg, "%s%s", pr, space);
		full_getkey(msg, 0);
		if (root_keys[key] == (short) quoted_insert) {
			mention("%s<Quote> ", msg);
			getkey();
			user_abort = 0;
		}
		check_abort();
		show_char(space, key);
		show_char(spaceless, key);
		index = (tab[key] == 0) ? alttab[key] : tab[key];
		if (index <= 0 || name_type(index) != NT_TABLE)
			break;
		alttab = index_table((tab[key] && !alttab[key]) ?
						tab[key] : alttab[key]);
		tab = index_table(index);
		if (!alttab)
			alttab = tab;
		strcat(space, " ");
	}
	say("");
	return ((tab[key] == 0) ? alttab : tab) + key;
}

char *temp_buf()	/* create a unique temporary buffer, return name */
{
	return bufnum_to_name(tmp_buf());
}

tmp_buf()	/* create a unique temporary buffer, return number */
{
	char bname[80];
	int uniq = 1;

	do {
		sprintf(bname, "-temp%d", uniq++);
	} while (exist(bname));
	return create(bname);
}

int strtoi(s, base)	/* convert a string of digits to a number */
char *s;
{
	char *p, *digits = "0123456789ABCDEF";
	int number, neg = 0;

	while (*s == '-')
		s++, neg = !neg;
	number = 0;
	if (!*s || !index(digits, toupper(*s)))
		got_bad_number = 1;
	while (*s && (p = index(digits, toupper(*s++))) && p - digits < base)
		number = number * base + (p - digits);
	return neg ? -number : number;
}

numtoi(s)	/* like strtoi(), but get base from string */
char *s;
{
	int neg = 0, i;

	got_bad_number = 0;
	while (isspace(*s))
		s++;
	if (*s == '\'')		/* allow 'x' as a number (give ascii code) */
		return s[1];
	while (*s == '-')
		s++, neg = !neg;
	if (s[0] == '0' && tolower(s[1]) == 'x')
		i = strtoi(s + 2, 16);
	else if (s[0] == '0' && tolower(s[1]) == 'o')
		i = strtoi(s + 2, 8);
	else if (s[0] == '0' && tolower(s[1]) == 'b')
		i = strtoi(s + 2, 2);
	else
		i = strtoi(s, 10);
	return neg ? -i : i;
}
