/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1993 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

#include "eel.h"

/*
Automatic indentation for C and EEL code.  Several indenting styles
are supported, selected by the following variables.

			*	*	*	*

The position of the closing brace is controlled by the Closeback
variable, 1 by default:

Closeback = 0;			Closeback = 1;
if (foo) {			if (foo) {
	bar();				bar();
	baz();				baz();
	}			}

By placing the opening brace on the following line, these styles
may be also used:

Closeback = 0;			Closeback = 1;
if (foo)			if (foo)
	{			{
	bar();				bar();
	baz();				baz();
	}			}


			*	*	*	*

The Topindent variable controls whether top-level statements in a function
are indented.  It is 1 by default.

Topindent = 0;			Topindent = 1;
foo()				foo()
{				{
if (bar)				if (bar)
	baz();					baz();
}				}

			*	*	*	*

The Matchdelim variable controls whether typing ), ], or } displays the
corresponding (, [, or { when the former is typed.  It uses the
show-matching-delimiter command.
*/

keytable c_tab;			/* key table for c mode */

	/* possible return values of in_c_comment */
#define IN_COMMENT	0x100	// general mask for any comment
#define IN_OLD_COMMENT	0x101	// in a /* */-type comment
#define IN_NEW_COMMENT	0x102	// in a //-type comment
#define IN_STRING	0x200	// general mask for any string-like thing
#define IN_STR_CONST	0x201	// inside double quotes
#define IN_CHAR_CONST	0x202	// inside single quotes

c_indenter()		/* replace surrounding indentation with new */
{
	if (in_c_comment(c_look_back) & IN_COMMENT)
		indent_c_comment();
	else
		indent_to_column(c_compute_indent());
}

char _c_mode_name[] = "C";

command c_mode()
{
	mode_keys = c_tab;		/* use these keys */
	c_tab[')'] = c_tab[']'] =
		Matchdelim ? (short) show_matching_delimiter : 0;
	indenter = c_indenter;
	auto_indent = 1;
	major_mode = _c_mode_name;
	strcpy(comment_start, "(/</|*>)[ \t]*");
	strcpy(comment_pattern, "//.*$|/<*>(.|<newline>)*<*>/<FirstEnd>");
	if (new_c_comments) {
		strcpy(comment_begin, "// ");
		strcpy(comment_end, "");
	} else {
		strcpy(comment_begin, "/* ");
		strcpy(comment_end, " */");
	}
	try_calling("c-mode-hook");
	if (c_mode_mouse_to_tag)
		mouse_goes_to_tag = 1;
	make_mode();
}

c_mode_hook()
{
}

/* make this the default mode for .c, .h, .e, etc. files */
suffix_c()	{ c_mode(); }
suffix_h()	{ c_mode(); }
suffix_e()	{ c_mode(); }
suffix_y()	{ c_mode(); }
suffix_cpp()	{ c_mode(); }
suffix_cxx()	{ c_mode(); }
suffix_hpp()	{ c_mode(); }
suffix_hxx()	{ c_mode(); }

jmp_buf c_indent_giveup;	/* come here when we hit narrowed edge */


/*  Look at the last two lines and return the correct indentation
    for the current line, assuming C source (or similar).  At start,
    we must be at the end of a line's indentation.
*/

c_compute_indent()
{
	int ind = 0;		/* indentation to use */
	int orig = point;
	int first_char = curchar(); /* cur line's first char */
	int pstart;		/* start position of prev line */
	int pafter;		/* after label on prev line */
	int pend;		/* end position of prev line */
	int prev_end = ';';	/* last char on previous line */
	int norm_narrow;	/* narrowing setting for most parsing */

	save_var point;
	if (setjmp(&c_indent_giveup))	/* if we hit end & gave up, */
		return ind;		/* use best guess so far */
	norm_narrow = MAX(narrow_start, point - c_look_back);
	save_var narrow_start = norm_narrow;
	if (first_char == '#')
		return 0;
	else if (first_char == '}')
		return cind_close();

	narrow_start = MAX(narrow_start, point - c_look_back / 10);
	to_begin_line();		/* don't look far back */
	cind_space();			/* skip whitespace and comments */
	prev_end = character(point - 1); /* get last relevant char */
	pend = point;

	to_statement_start();		/* if contin line, back up */
	pstart = point;			/* now at start of biggest statement */
	past_labels();			/* containing original point */
	pafter = point;			/* just after label */
	ind = current_column();
	narrow_start = norm_narrow;	/* ok to look far back again */

	if (is_case(orig)) {
		ind = indent_case(orig, pafter);
	} else if (have_label(orig)) {
		ind = c_label_indent;
	} else if (prev_end == ':' && is_label(pstart)) {
		ind = get_column(pstart);
		if (!(Closeback && just_open(orig)))
			ind += get_c_indent();
	} else if (is_else(orig)) {
		point = orig;
		ind = indent_else(ind);
	} else if (prev_end == ';') {
		ind = indent_after_statement();
	} else if (prev_end == '}') {
		point = pend;
		ind = indent_after_block();
	} else if (prev_end == '{') {
		if (!ind)	/* special case if no previous indentation */
			return ind + top_indent(prev_end) ? get_c_indent() : 0;
		if (Closeback || curchar() != '{')
			ind += get_c_indent();
		if (just_open(pafter))
			ind -= c_brace_offset;
	} else if (is_contin_statement(pend)) {
		return indent_c_contin(ind, orig, pend);
	} else if (!index(":);{}", c_line_before()) && !c_statement_start()
				&& !(Closeback && just_open(point)))
		return ind;
	else if (just_open(orig))
		return ind + c_brace_offset + (Closeback ? 0 : get_c_indent());
	else
		ind += get_c_indent();
	return ind;
}

cind_search(pat)	/* search back for pattern */
char *pat;		/* if we reach start of (examinable) buffer, give up */
{			/* call this only from c_compute_indent */
	int res = re_search(-1, pat);

	if (point == narrow_start)
		longjmp(&c_indent_giveup, 1);
	return res;
}

cind_space()		/* move back over any whitespace or comments */
{			/* give up if we hit start */
	if (cind_try_space())
		longjmp(&c_indent_giveup, 1);
}

cind_try_space()	/* move back over whitespace, comments, preproc */
{			/* lines, and return 0, or 1 if we hit start */
	for (;;) {
		if (index(" \t\n", character(point - 1)))
			point--;
		else if (cind_skip_new_comment() || cind_skip_preproc_line())
			;
		else if (character(point - 1) == '/'
					&& character(point - 2) == '*') {
			point -= 2;
			if (!search(-1, "/*") || point == narrow_start)
				return 1;
		} else
			return point == narrow_start;
	}
}

cind_skip_new_comment()	/* move before // comment on line */
{			/* return 1 if there was one */
	int old = point;

	save_var narrow_start = give_begin_line();
	while (search(-1, "//"))
		if (at_c_comment(point + 2, 0) == IN_NEW_COMMENT)
			return 1;
	point = old;
	return 0;
}

cind_skip_preproc_line()	/* move back past any preproc line */
{				/* return 1 if there was one */
	int old = point;

	for (;;) {
		to_begin_line();
		if (character(point - 2) != '\\')
			break;
		point -= 2;	/* prev might be preproc contin line */
	}
	if (parse_string(1, "[ \t]*#"))
		return (point > narrow_start);
	point = old;
	return 0;
}

c_line_before()		/* retrieve last character of prev line of c code */
{			/* (ignoring comments, blank lines, etc.) */
	save_var point;		/* assumes we're in this line's indentation */
	return cind_try_space() ? ';' : character(point - 1);
}

/*  Look at the last two lines and guess at the correct indentation,
    assuming C source (or similar).  If we're not in this line's
    indentation, though, or our new indentation matches the old,
    just indent one more level.
*/

command do_c_indent() on c_tab['\t']
{
	int orig = point, ind = get_c_indent();
	int orig_column = current_column();

	to_indentation();
	if (!c_tab_always_indents && orig_column > current_column()) {
					/* if not in indentation */
		point = orig;
		to_column(orig_column + ind - orig_column % ind);
	} else if (prev_cmd == C_INDENT)	/* repeated, make bigger */
		to_column(orig_column + ind);
	else
		c_indenter();
	this_cmd = C_INDENT;
}

/*
Tell if more indent is required.
Assumes last line had no indent and point is at its start.
The hard case is distinguishing between these when Topindent is zero:

	foo();				foo();
}				}
func(bar)			if (bar)

When point is just after these examples, we must examine the word
at "func" to distinguish the left example (the end of a function and
beginning of the next, requiring no indentation) from the right (the
end of a block and start of a conditional, requiring indentation).
*/
top_indent(prev_end)
{
	if (Topindent)
		return (prev_end == '{');
	switch (prev_end) {
		case ';':
		case '}':	return 0;
		default:	return 1;
		case '{':	if (parse_string(1, "[ \t]*{")) {
					cind_try_space();
					to_indentation();
				}
		case ')':	break;
	}
	return c_statement_start();
}

/*
Tell if the last line begins a statement (rather than a function
declaration or the continuation of a previous statement).
Assumes point is at the start of the last line.
*/
c_statement_start()
{
	return parse_string(1, "[{} \t]*(if|else|while|do|for)[^a-zA-Z0-9]");
}

just_open(p)
{
	save_var point = p;
	return parse_string(1, "[ \t]*{[ \t]*($|//|/<*>)");
}

is_label(p)	/* does line at p contain only case: or label? */
{
	save_var point = p;
	while (parse_label())	/* go past any labels */
		point = matchend;
	return parse_string(1, "[ \t]*\n|/<*>|//");
}

is_case(p)	/* is there a case-like statement at position p? */
{
	if (!index("cdp", character(p)))
		return 0;
	save_var point = p;
	return parse_string(1, "(case|default|public|private|protected)"
			    "[^a-zA-Z0-9_]");
}

indent_case(orig, p)	// indent case statement at orig, using prev line at p
{
	if (is_switch(p) || just_open(p) && is_switch(p = last_c_line(p))) {
		point = p;		/* indent case after switch */
		return current_column() + get_c_indent() + c_case_offset;
	}
	point = orig;		/* copy indentation from prior case */
	return cind_find_case();
}

is_switch(p)	/* is there a switch statement at position p? */
{
	if (!index("sc", character(p)))
		return 0;
	save_var point = p;
	return parse_string(1, "(switch|class)[^a-zA-Z0-9_]");
}

is_else(p)	/* is there an else keyword at position p? */
{
	if (character(p) != 'e')
		return 0;
	save_var point = p;
	return parse_string(1, "else[^a-zA-Z0-9_]");
}

last_c_line(p)	/* return pos of start of c line before line with p */
{
	save_var point = p;
	to_begin_line();
	cind_space();
	to_statement_start();
	return point;
}

in_c_comment(lookback)		// is point inside a c comment, string, etc?
{
	int i = scan_for_string();

	if (i || !lookback)	// check this line first
		return i;
	save_var narrow_start = point - lookback;
	save_var point;
	while (re_search(RE_REVERSE | RE_FASTEST, "/<*>|<*>/")) {
		point++;
		i = scan_for_string();
		if (!i || i == IN_OLD_COMMENT)	// did we find "/*"?
			return curchar() == '*' ? IN_OLD_COMMENT : 0;
	}
	return 0;		// not found, assume we're not in comment
}

scan_for_string()	// check current line to see if we're in
{			// string or new-style comment
	int c;

	save_var point;
	save_var narrow_end = size() - point;
	to_begin_line();		// scan this line before point
	while (re_search(1, "<squote|dquote>|//|/<*>")) {
		point = matchstart;
		c = curchar();
		if (c == '/' && character(point + 1) == '/')  // found "//"
			return IN_NEW_COMMENT;
		else if (!c_skip_item(c, RE_FORWARD))
			return c == '/' ? IN_OLD_COMMENT
			     : c == '"' ? IN_STR_CONST : IN_CHAR_CONST;
	}
	return 0;
}

at_c_comment(pos, lookback)
{
	save_var point = pos;
	return in_c_comment(lookback);
}

indent_c_comment()		/* indenter when inside a comment */
{
	int prev_indent;
	int orig, has_star, no_star = 0;

	has_star = parse_string(1, "[ \t]*<*>");
	orig = point;
	re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
	to_indentation();
	if (has_star && parse_string(1, "[/ \t]*<*>"))
		point = matchend - 1;	// line up star under prev star
	else
		no_star = 1;
	prev_indent = current_column();
	point = orig;
	if (no_star && parse_string(1, "<*>/"))
		prev_indent = 0;
	to_column(prev_indent);		/* indentation as previous */
}

/* We are at the beginning of a statement, which may be part of a larger
 * statement because it has "if ()" or similar before it.  Find the
 * biggest statement that the current line could be the tail end of,
 * and return its indentation.  We assume the statement starts at the
 * beginning of a line. */

indent_after_statement()
{
	int lastcont;

	for (;;) { /* was only the end of a larger statement, find start */
		do {	/* look back for line that ends with ;{}: */
			lastcont = point;
			to_begin_line();
		} while (!cind_try_space()
				&& !index(";{}:", character(point - 1)));
		point = lastcont;	/* use indentation of line */
		to_c_indentation();	/* just after it */
		if (!parse_string(RE_FORWARD, "(}[ \t]*)?e!lse[^a-zA-Z0-9_]"))
			break;
		point = matchend;
		if (!to_matching_if())
			break;
	}
	return current_column();
}

/* Return new indentation, when the previous line ended with a } character.
 * We are now just after the } character. */

indent_after_block()
{
	c_rev_skip_level('{', '}');	/* find matching { */
	to_begin_line();
	return indent_after_statement();
}

indent_c_contin(ind, orig, lastl)	/* indent continuation line */
{
	char c, plevel = 0, qlevel = 0, findq;

	if (!c_align_contin_lines)	// never line up
		return ind + get_c_indent() + c_contin_offset;
	to_begin_line();
	save_var narrow_start = point;
	point = orig;
	findq = (character(orig) == ':');
	while (re_search(-1, "[()?:\"'\n]|[=!<>]?=")) {
		switch (c = curchar()) {
			case '(':	/* indent after unmatched lparen */
				if (plevel-- > 0)
					break;
indhere:			point++;
				if (!parse_string(1, "[ \t]*/[*/]"))
					re_search(1, "[ \t]*");
				return limit_c_contin_indent(ind);
			case ')':
				plevel++;
				break;
			case '?':	/* indent : to matching ? */
				if (qlevel-- > 0 || plevel > 0 || !findq)
					break;
				point--;
				goto indhere;
			case ':':
				qlevel++;
				break;
			case '=': case '!':
				if (character(point + 1) == '=' || plevel > 0)
					break;
				goto indhere;
			case '\n':		 // skip over comments
				cind_try_space(); // at ends of lines
				break;
			case '"': case '\'':
				point++;
				c_skip_item(c, RE_REVERSE);
				break;
		}
	}
	if (character(lastl - 1) == ',') {
		point = lastl;
		to_begin_line();
		point += parse_string(1, "(.*[ \t]on)?[ \t]+");
		return limit_c_contin_indent(ind);
	}
	if (!ind)
		return 0;
	return ind + 2 * get_c_indent() + c_contin_offset;
}

limit_c_contin_indent(ind)	// if lining up under something would put us
{				// too far to right, just use std indent
	int std, i = current_column();	// get lining-up indent

	if (i > c_align_contin_lines) {
		std = ind + get_c_indent() + c_contin_offset;
		i = MIN(std, i);
	}
	return i;
}

indent_else(def)	/* indent a line that starts with "else" */
{			/* by finding matching "if" */
	if (!to_matching_if())
		return def;
	to_c_indentation();
	return current_column();
}

to_matching_if()	/* go to matching if statement, assuming we're */
{			/* just before an else keyword */
	int extra_elses = 0;	/* ret 0 if couldn't find it, 1 if ok */

	while (cind_search("<min>(([^a-zA-Z0-9]|^)(else|if)[^a-zA-Z0-9]|})")) {
		if (!in_c_comment(2000)) {
			if (curchar() == '}') {
				point++;	/* find matching { */
				c_rev_skip_level('{', '}');
			} else if (character(point + 1) == 'e')
				extra_elses++;
			else if (--extra_elses < 0)
				return 1;
		}
	}
	return 0;
}

/* Find a prior case statement, and return its indentation. */

cind_find_case()
{
	do {
		if (!cind_search("[^a-zA-Z0-9_](case|default|public|private)"
						"[^a-zA-Z0-9_]"))
			break;
	} while (in_c_comment(2000));
	return get_indentation(point);
}

parse_label()	/* parse for a case:, default: or label construct at point */
{		/* return nonzero if found, setting matchend to end */
	return have_colon() && parse_string(1,
		"(case[^a-zA-Z0-9_\n]([^:\n]*|.*':'.*)"
		"|[a-zA-Z0-9_]+[ \t]*):[ \t]*![^:]");
}

have_colon()		/* is there a colon later on this line? */
{
	save_var narrow_end = size() - give_end_line();
	save_var point;
	return search(1, ":");
}

have_label(pt)	/* is there a label here? */
{
	save_var point = pt;
	return have_colon() && parse_string(1,"[a-zA-Z0-9_]+[ \t]*:[^:]");
}

to_c_indentation()	/* go to end of indentation, ignoring any */
{			/* case : or label-like constructs */
	to_indentation();
	past_labels();
}

past_labels()	/* move past labels or case statements at point */
{
	while (parse_label() && character(matchend) != '\n')
		point = matchend;
}

/* If this line appears to be a continuation line, back up until we find
 * one that doesn't.  (We assume it's a continuation line if the previous
 * line doesn't end with one of the five characters ; { } ) : or a
 * keyword like "else".)  Then go to end of indentation.
*/

to_statement_start()
{
	int lastcont, count = 10;	/* give up after this many lines */

	to_begin_line();
	if (parse_string(1, "[ \t]*}")) {
		search(1, "}");
		c_rev_skip_level('{', '}');
	}
	do {
		to_begin_line();
		lastcont = point;
	} while (!cind_try_space() && is_contin_statement(point)
			&& count-- > 0);
	point = lastcont;
	to_indentation();
}     

#if 0

is_contin_statement(p)	/* would line after p be a continuation line? */
{			/* assumes p is just after last interesting char */
	switch (character(p - 1)) {
		case ';':
		case '{':
		case '}':
			return 0;
		case ')':
			save_var point = p;
			save_var narrow_start = p - 500;
			if (!c_rev_skip_level('(', ')'))
				return 1;
			return !parse_string(-1,
				"[^a-z0-9_](if|while|for|switch)[ \t]*");
		case 'e':
			return !is_else(p - 4);
		case 'o':
			save_var point = p;
			return !parse_string(-1, "[^a-z0-9_]do");
		case ':':
			save_var point = p;
			to_indentation();
			return !is_label(point);
		default:
			return 1;
	}
}

#else

is_contin_statement(p)	/* would line after p be a continuation line? */
{			/* assumes p is just after last interesting char */
	switch (character(p - 1)) {
		case ';':
		case '{':
		case '}':
			return 0;
		case ')':
			save_var point = p;
			save_var narrow_start = p - 500;
			if (!c_rev_skip_level('(', ')'))
				return 1;
           		if (parse_string(-1, "^[a-z_]+.*"))
	                        return 0;
                        return !parse_string(-1,
				"[^a-z0-9_](if|while|for|switch)[ \t]*");
		case 'e':
			return !is_else(p - 4);
		case 'o':
			save_var point = p;
			return !parse_string(-1, "[^a-z0-9_]do");
		case ':':
			save_var point = p;
			to_indentation();
			return !is_label(point);
		default:
			return 1;
	}
}


#endif

cind_close()	/* indent a line that starts with a } character */
{		/* find matching { and indent based on that one */
	int ind, solitary;

	point++;		/* go past } char */
	c_rev_skip_level('{', '}');	/* find matching { */
	solitary = parse_string(-1, "\n[ \t]*");	/* stuff before? */
	if (solitary)
		to_indentation();
	else
		to_statement_start();
	past_labels();
	ind = current_column();
	if (!Closeback && !solitary)
		ind += get_c_indent() + c_brace_offset;
	return ind;
}

	/* dquote "(" snormal OR bslash anything ")" STAR dquote */
#define C_STRING "\"([^\"\\\n]|\\(.|\n))*\""

	/* squote "(" cnormal OR bslash anything ")" STAR squote */
#define C_CCONST "'([^'\\\n]|\\(.|\n))*'"

/* Skip past a string, character-constant, or comment,
   depending on "item".
*/

c_skip_item(item, flags)
{
	char *s; int m;

	switch (item) {
		case '/':	m = (flags & RE_REVERSE) ? -1 : 1;
				return search(m, (m < 0) ? "*/" : "/*")
					&& search(m, (m < 0) ? "/*" : "*/");
		case '"':	s = C_STRING; break;
		case '\'':	s = C_CCONST; break;
	}
	return re_search(flags, s);
}

/* look backwards for a matching pair of delimiters, ignoring
 * delimiters inside strings, character constants, or comments
 */

int c_rev_skip_level(left_d, right_d)
char left_d, right_d;
{
	char pattern[40], ch;
	int level = 0;

	sprintf(pattern, "<*>/|[%c%c'\"]|//.*\n", right_d, left_d);

bgn:	do {
		if (!re_search(RE_REVERSE, pattern))
			return 0;
		switch (ch = curchar()) {
			case '/':
				if (at_c_comment(point + 2, 100) & IN_STRING)
					to_end_line();
				break;
			case '\'': case '"': case '*':
				point = matchstart;
				c_skip_item(character(point - 1), RE_REVERSE);
				goto bgn;
			default:
				if (ch == right_d)
					level++;
				else if (ch == left_d)
					level--;
		}
	} while (level);
	return 1;
}

fix_c_indentation()		/* recompute this line's indentation */
{				/* without moving point */
	save_spot point;
	to_indentation();
	c_indenter();
}


/*  fix indentation if necessary when { is typed */
command c_open() on c_tab['{']
{
	normal_character();
	if (Closeback && current_column() - 1 <= get_indentation(point)
			&& !in_c_comment(c_look_back))
		fix_c_indentation();
}

/*  fix indentation if necessary when } is typed */
command c_close() on c_tab['}']
{		/* only if typed inside indentation, & it might need fixup */
	normal_character();
	if (current_column() - 1 <= get_indentation(point)
			&& !in_c_comment(c_look_back))
		fix_c_indentation();
	if (Matchdelim)
		find_delimiter();
}

/*  fix indentation if necessary when : is typed */
command c_colon() on c_tab[':']
{
	normal_character();
	if (parse_string(-1, "^[ \t]*(case[^a-zA-Z0-9_\n][^:\n]*"
				"|[a-zA-Z0-9_]+[ \t]*):")
			&& !in_c_comment(c_look_back))
		fix_c_indentation();
}

/*  fix indentation if necessary when # is typed */
command c_hash_mark() on c_tab['#']
{
	normal_character();
	if (parse_string(-1, "^[ \t]*#"))
		fix_c_indentation();
}

get_c_indent()	/* indent this many spaces for each level */
{			/* if c_indent is 0, use tab_size instead */
	return (c_indent > 0) ? c_indent : tab_size;
}
