/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

#include "eel.h"

buffer char _is_line = 0;

command set_show_graphic() on reg_tab[FCTRL(6)]
{	/* cycle through the three methods of displaying ctrl & meta chars */
	int mode, i, std;

	if (!has_arg) {	/* determine old mode */
		switch (_display_class[0]) { /* check value for null char */
			case BNORMAL: mode = 1; break;
			case BHEX: mode = 2; break;
			default: mode = 0;
		}
		mode = (mode + 1) % 3;	/* get to next mode */
	} else
		mode = iter % 3;
	iter = 1;
	std = (mode == 0) ? BM : (mode == 1) ? BNORMAL : BHEX;
	for (i = 0; i < 128; i++)
		if (i < 32 || i == 127) {
			if (i != '\n' && i != '\t')
				_display_class[i] = mode ? std : BC;
			_display_class[i + 128] = mode ? std : BMC;
		} else {
			_display_class[i] = BNORMAL;
			_display_class[i + 128] = std;
		}
	full_redraw = 1;
}

command set_tab_size()
{
	tab_size = get_number("Tab size: ");
	if (tab_size < 2 || tab_size > 256)
		tab_size = 8;
	say("Tab size is %d in this buffer.", tab_size);
	build_first = 1;
}

maybe_ding(want) /* ring if we want this type of bell & want bell generally */
{
	if (want && want_bell)
		ding();
}

fix_cursor()
{
	cursor_shape = over_mode ? overwrite_cursor : normal_cursor;
}

make_mode()	/* tell eps to rebuild mode line for each window showing */
{		/* this buf */
	must_build_mode = 1;
	fix_cursor();
}

prepare_windows()	/* fix up borders before each window */
{			/* is displayed */
	int i = -1, j = BORD(MODE_EDGE, -1), popup, wantmode;

	popup = (is_window(window_handle) == ISPOPUP);
	_window_flags &= ~WANT_MODE_LINE;
	wantmode = (popup ? (_window_flags & FORCE_MODE_LINE)
		: !(_window_flags & NO_MODE_LINE));
	if (wantmode)
		_window_flags |= WANT_MODE_LINE;
	if (!popup) {
		if (!window_left && border_left)
			j |= BORD(BLEFT, -1);
		if (!window_top && border_top)
			j |= BORD(BTOP, -1);
		if (window_left + window_width >= screen_cols ? border_right
							      : border_inside)
			j |= BORD(BRIGHT, -1);
		if (window_top + window_height >= screen_lines
							    && border_bottom)
			j |= BORD(BBOTTOM, -1);
		if (!(_window_flags & FIXED_BORDERS))
			display_scroll_bar = tiled_scroll_bar;
		i = tiled_border & j;
	} else if (wantmode) {
		i = get_wattrib(window_number, PBORDERS);
		if (!(i & BORD(MODE_EDGE, -1)))
			i |= tiled_border & BORD(MODE_EDGE, -1);
	}
	if (i != -1 && !(_window_flags & FIXED_BORDERS))
		set_wattrib(window_number, PBORDERS, i);
}

build_mode()	/* construct mode line for this window */
{
	char mode_line[FNAMELEN];

	if (_window_flags & WANT_MODE_LINE) {
		assemble_mode_line(mode_line);
		window_title(window_number, MODE_EDGE, TITLELEFT(2),
			mode_line);
	}
}

assemble_mode_line(m)	/* construct mode line in m */
char *m;
{
	int i, left;
	char *p, file[FNAMELEN];

	strcpy(m, mode_start);
	p = m + strlen(m);
	if (*filename)
		relative(filename, file);
	left = ptrlen(p) - 30;
	if (!*filename || fnamecmp(get_tail(filename, 0), bufname)) {
		if (is_dired_buf())
			sprintf(p, "%.*s ", left, bufname);
		else
			sprintf(p, "(%.*s) ", left, bufname);
		p += strlen(p);
		left = ptrlen(p) - 30;
	}
	if (*filename)			/* if file, put name */
		sprintf(p, "%.*s ", left, file);
	for (i = _recursion_level; i-- >= 0;)
		strcat(p, "[");
	set_mode(p + strlen(p));	/* build part inside []'s in mode */
	for (i = _recursion_level; i-- >= 0;)
		strcat(p, "]");
	strcat(p, mode_end);
}

set_mode(mode)	/* add major and minor modes to mode line */
char *mode;
{
	strcpy(mode, major_mode);
	if (fill_mode)
		strcat(mode, " Fill");
	if (over_mode)
		strcat(mode, " Over");
	if (in_rectangle_mode)
		strcat(mode, " Rect");
	if (strip_returns != strip_returns.default)
		strcat(mode, strip_returns ? " Trans" : " NoTrans");
	if (_len_def_mac)
		strcat(mode, " Def");
	if (narrow_start > 0 || narrow_end > 0)
		strcat(mode, " Narrow");
	if (_buf_readonly)
		strcat(mode, " RO");
}

display_more_msg(w)	/* put up "More" in lower right corner */
{
	window_title(w, BBOTTOM, TITLERIGHT(1), "%m");
}

command show_version()
{
	say("Epsilon Programmer's Editor %s", version);
}

command highlight_region() on cx_tab[CTRL('H')]	/* toggle highlighting */
{
	if (has_arg ? (iter != 0) : !is_highlight_on())
		highlight_on();
	else
		highlight_off();
	iter = 1;
}

is_highlight_on()	/* 0 if no, else 1 region, 2 rect, 3 line highlight */
{
	return _highlight_control ? modify_region(1, MRTYPE, -1) : 0;
}

#if 0

highlight_on()
{
	modify_region(1, MRTYPE, _is_rectangle ? REGRECT : REGNORM);
	_highlight_control = 1;
	call_on_modify = 1;
}

#else

highlight_on()
{
	modify_region(1, MRTYPE,
                  _is_rectangle ? REGRECT : _is_line ? REGLINE : REGNORM);
	_highlight_control = 1;
	call_on_modify = 1;
}      


#endif

highlight_off()
{
	_highlight_control = 0;
}

command redisplay()
{
	refresh();
}
