/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

#include "eel.h"

buffer char _is_line;

/*  Make this line's indentation the same as the previous nonempty line's,
    unless the indentation would be less, or we're not in this line's
    indentation.  In those cases, indent to the next tab stop. 
*/

command indent_previous() on reg_tab[CTRL('I')]
{
	int orig_column, prev_indent;
	int orig = point;

	orig_column = current_column();	/* point's column */
	to_begin_line();
	re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
	to_indentation();
	prev_indent = current_column();
	point = orig;
	to_indentation();		/* go to current line's indent */

	if (orig_column < prev_indent && orig_column <= current_column())
		indent_to_column(prev_indent);	/* indentation as previous */
	else {
		point = orig;
		indent_like_tab();
	}
}

prev_indenter()		/* indent this line like previous one */
{
	int prev_indent;
	int orig;

	to_begin_line();
	orig = point;
	re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
	to_indentation();
	prev_indent = current_column();
	point = orig;
	indent_to_column(prev_indent);	/* indentation as previous */
}


/*  Using the previous non-empty line, "tab" over to the beginning
    of the next column of darkspace.
*/

command indent_under() on reg_tab[ALT(CTRL('I'))]
{
	int prev_indent, indent, orig = point;

	indent = current_column();
	to_begin_line();
	re_search(-1, "[^ \t\n]");	/* find previous non-blank line */
	move_to_column(indent);		/* go to same column on prev line */
	if (current_column() > indent)
		point--;

	if (!re_search(1, "[ \t\n]") || character(point - 1) == '\n' ||
			!re_search(1, "[^ \t]")
			|| character(point - 1) == '\n') {
		point = orig;			 /* indent like one tab */
		indent_like_tab();
	} else {				/* get new place to tab to */
		prev_indent = current_column() - 1;
		point = orig;			/* back to original line */
		indent_to_column(prev_indent);	/* force to new indentation */
	}
}

indent_like_tab()	/* indent like tab char, but maybe use spaces */
{
	int cur = current_column();

	to_column(cur + tab_size - cur % tab_size);
}

/* position point before the first darkspace character on the current line */
command to_indentation() on reg_tab[ALT('m')]
{				/* used by other commands */
	to_begin_line();
	re_search(1, "[ \t]*");
}

to_column(place) /* replace current indentation with indent to column place */
{
	delete(point - parse_string(-1, "[ \t]*"), point); /* del old ind */
	insert_to_column(current_column(), place);
}

indent_to_column(place)	 /* delete whitespace, then indent */
{
	delete_horizontal_space();
	insert_to_column(current_column(), place);
}

insert_to_column(start, end)	/* insert to get from start to end columns */
{
	int i;

	if (indent_with_tabs)
		while ((i = start + tab_size - start % tab_size) <= end) {
			insert('\t');
			start = i;
		}
	for (; start < end; start++)
		insert(' ');
}

force_to_column(want)	/* convert tabs to spaces or add indentation to */
{			/* get to this column on current line */
	int have;

	move_to_column(want);
	have = current_column();
	if (have == want)
		return;
	if (point >= size() - narrow_end || curchar() == '\n')
		insert_to_column(have, want);
	else {
		hack_tabs(want - have);
		move_to_column(want);
	}
}

get_column(pos)		/* give column of a buffer position */
{
	save_var point = pos;
	return current_column();
}

column_to_pos(col)	// give buffer position of char at column on this line
{
	save_var point;
	move_to_column(col);
	return point;
}

get_indentation(pos)	/* give indentation of line at buffer position */
{
	save_var point = pos;
	to_indentation();
	return current_column();
}

command indent_rigidly() on cx_tab['\t']
{
	int temp, indent;

	if (!has_arg) iter = tab_size;
	if (point > mark) {
		temp = mark;
		mark = point;
		point = temp;
	}
	if (nl_reverse())	/* to start of line */
		point++;
	temp = point;
	while (point < mark) {
		re_search(1, "[ \t]*");
		indent = current_column() + iter;
		to_column(indent);
		if (curchar() == '\n')
			delete_horizontal_space();
		nl_forward();
	}
	mark = temp;
	iter = 0;
}

command indent_region() on reg_tab[ALT(CTRL('\\'))]
{
	int temp;
	int (*func)() = (int (*)()) (mode_keys['\t'] ? mode_keys['\t']
						: root_keys['\t']);

	if (point > mark) {
		temp = mark;
		mark = point;
		point = temp;
	}
	if (nl_reverse())
		point++;
	temp = point;
	while (point < mark) {
		(*func)();
		if (curchar() == '\n')
			delete_horizontal_space();
		nl_forward();
		check_abort();
	}
	mark = temp;
	this_cmd = iter = 0;
}

hack_tabs(offset)	/* change tabs to spaces between current column */
{			/* and current + offset column */
	int count = offset, dir = 1;
	spot orig;

	if (offset < 0)
		count = -offset, dir = -1;
	if (count > 1 || character(point - (offset < 0)) == '\t') {
		orig = alloc_spot();
		column_chars(count, dir);
		tab_convert(point, *orig, 0);
		point = *orig;
		free_spot(orig);
	}
}

column_chars(n, dir)		/* move by n chars, counting tabs as spaces */
{
	int i, j, sofar = 0, failed = 0;
	int offs;

	offs = (dir == -1) ? -1 : 0;
	while (sofar < n && point + offs >= narrow_start
				&& point + offs < size() - narrow_end) {
		if (character(point + offs) == '\n')
			point += dir, sofar++;
		else {
			i = current_column();
			if (failed)
				point += dir;
			else
				move_to_column(i + (n - sofar) * dir);
			j = current_column();
			failed = (i == j);
			sofar += (j - i) * dir;
		}
		check_abort();
	}
}

all_blanks(from, to)	/* are all chars in this region whitespace? */
{
	save_var point;
	for (point = from; point < to; point++)
		if (!isspace(curchar()))
			return 0;
	return 1;
}

command untabify_region() on cx_tab[ALT('i')]
{
	do_tabify_region(0);
}

command tabify_region() on cx_tab[ALT(CTRL('i'))]
{
	do_tabify_region(1);
}

#if 0

do_tabify_region(totabs)	/* run tab-convert on current region */
{
	tab_convert(point, mark, totabs);
	iter = 0;
	say("Done.");
}

#else

do_tabify_region(totabs)    /* run tab-convert on current region */
  {
    int   StartCol;
    int   EndCol;
    int   temp;
    save_spot mark;
    save_spot point;

    if (_is_rectangle)
      {
        block_setup(&StartCol, &EndCol);
        move_to_column(StartCol);
    
        while (point <= mark)
          {
            temp = point;
            move_to_column(EndCol);
            tab_convert(temp, point, totabs);
            point = temp;
            nl_forward();
            move_to_column(StartCol);
          }
      }
    else
      {
        if (_is_line)
            line_up_mark_and_point();
        tab_convert(mark, point, totabs);
      }
    iter = 0;
    say("Done.");
  }

#endif

tab_convert(from, to, totabs)
{			/* convert between tabs & spaces in this region */
	spot orig = alloc_spot(), final = alloc_spot();
	int abort = 0, old_tabs = indent_with_tabs;
	jmp_buf *old_level = top_level, this_level;

	point = from;
	if (from > to)
		point = to, to = from;
	*final = to;
	top_level = &this_level;
	indent_with_tabs = 1;	/* use tabs, even if we normally don't */
	if (setjmp(top_level))
		abort = 1;
	else {
		if (totabs)
			spaces_to_tabs(final);
		else
			tabs_to_spaces(final);
		point = *orig;
	}
	free_spot(orig);		/* free these, even if aborted */
	free_spot(final);
	top_level = old_level;
	indent_with_tabs = old_tabs;
	if (abort)
		quick_abort();
}

tabs_to_spaces(final) /* convert tabs to spaces from point to *final */
spot final;
{
	int i, j;

	while (search(1, "\t") && point <= *final) {
		i = current_column();
		point--;
		j = current_column();
		replace(point, ' ');
		while (++j < i)
			insert(' ');
	}
}

#if 0

spaces_to_tabs(final) /* convert spaces to tabs from point to *final */
spot final;
{
	int start, temp, i, j;

	while (point < *final) {
		if (!re_search(1, "(\t| [ \t])[ \t]*"))
			break;
		temp = point;
		point = start = matchstart;
		i = current_column();
		point = temp;
		if (point > *final)
			point = *final;
		j = current_column();
		if (i + 1 < j) {
			delete(start, point);
			insert_to_column(i, j);
		}
	}
}

#else

spaces_to_tabs(final) /* convert spaces to tabs from point to *final */
spot final;
{
    int start, temp, i, j;

    while (point < *final) {
        if (!re_search(1, "(\t| [ \t])[ \t]*"))
            break;
        temp = point;
        point = start = matchstart;
        i = current_column();
        point = temp;
        if (point > *final)
          {
            point = *final;
            if (point <= start)
                break;
          }
        j = current_column();
        if (i + 1 < j) {
            delete(start, point);
            insert_to_column(i, j);
        }
    }
}

#endif

