/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

#include "eel.h"

check_num_kill_buffers()
{
	int i;

	if (kill_buffers == _prev_kill_buffers && _kill_width)
		return;
	if (kill_buffers < 1)
		kill_buffers = 1;
	if (_kill_width)
		free(_kill_width);
	_kill_width = (short *) malloc(kill_buffers * sizeof(short));
	for (i = 0; i < _prev_kill_buffers; i++) {
		sprintf(_cur_kill_buf, "-kill-buffer-%d", i);
		delete_buffer(_cur_kill_buf);
	}
	for (i = 0; i < kill_buffers; i++)
		_kill_width[i] = UNUSED;
	_prev_kill_buffers = kill_buffers;
	_num_kill = 0;
}

/* Prepare to use another kill buffer.  Find the proper one and empty it. */
push_kill()
{
	_num_kill = (++_num_kill) % kill_buffers;
	sprintf(_cur_kill_buf, "-kill-buffer-%d", _num_kill);
	zap(_cur_kill_buf);	/* zap will create if necessary */
}

/* make the previous kill buffer current, create if necessary */
pop_kill()
{
	check_num_kill_buffers();
	if (--_num_kill < 0)
		_num_kill += kill_buffers;
	sprintf(_cur_kill_buf, "-kill-buffer-%d", _num_kill);
	create(_cur_kill_buf);	/* no effect if already exists */
}

/* transfer region to kill buffer appropriately, then delete it */
do_save_kill(near, far)		/* near is end adjacent to old text */
{
	if (near != far) {
		do_save_only(near, far);
		delete(near, far);
	}
}

/* transfer region to kill buffer appropriately */
do_save_only(near, far)		/* near is end adjacent to old text */
{
	char *thisbuf = bufname;

	check_num_kill_buffers();
	if (prev_cmd != APPENDKILL || _kill_width[_num_kill] != REGION) {
		push_kill();		/* start on a new killbuffer */
		_kill_width[_num_kill] = REGION;
	} else { /* get to proper end of same kill buffer as last time */
		bufname = _cur_kill_buf;
		point = (near < far) ? size() : 0;
		bufname = thisbuf;
	}
	xfer(_cur_kill_buf, near, far);
	this_cmd = APPENDKILL;
}

/* exchange point and mark */
command exchange_point_and_mark() on cx_tab[CTRL('X')]
{
	int temp = point, tcol;

	point = mark;
	tcol = (mark == _mark_virtual_pos) ? _mark_virtual_column : -1;
	do_set_mark(temp);
	cursor_to_column = tcol;
}

/* set mark to be point */
command set_mark() on reg_tab[CTRL('@')], reg_tab[ALT('@')], reg_tab[ALT(' ')]
{
	do_set_mark(point);
	say("Mark set.");
	_is_rectangle = in_rectangle_mode;
	highlight_on();
}

do_set_mark(val)
{
	mark = val;
	_mark_virtual_column = cursor_to_column;
	_mark_virtual_pos = val; /* remember these for rect cmds */
}

kill_after_cmd(cmd)	/* kill from here to position after cmd */
int (*cmd)();
{
	int start;

	check_modify(bufnum);
	start = point;
	(*cmd)();
	do_save_kill(start, point);
}

/* delete from point to the end of the next word */
command kill_word() on reg_tab[ALT('d')]
{
	kill_after_cmd(forward_word);
}


/* delete from point to the beginning of the previous word */
command backward_kill_word() on reg_tab[ALT(CTRL('H'))],
			reg_tab[NUMALT(GREYBACK)]
{
	kill_after_cmd(backward_word);
}

command kill_sentence() on reg_tab[ALT('k')]
{
	kill_after_cmd(forward_sentence);
}

command kill_level() on reg_tab[ALT(CTRL('K'))]
{
	kill_after_cmd(forward_level);
}

command backward_kill_level() on reg_tab[NUMALT(KEYDELETE)]
{
	kill_after_cmd(backward_level);
}


/* delete the current line, and put it in the kill buffer */
command kill_line() on reg_tab[CTRL('K')]
{
	int start = point;

	check_modify(bufnum);
	if (has_arg) {
		if (iter > 0)
			while (iter-- > 0)
				nl_forward();
		else {
			while (iter++ <= 0)
				nl_reverse();
			if (point > 0)
				point++;
		}
	} else if (curchar() == '\n')
		point++;
	else
		to_end_line();

	iter = 0;
	do_save_kill(start, point);
}


command kill_region() on reg_tab[CTRL('W')]
{
	check_modify(bufnum);
	if (_is_rectangle)
		kill_rectangle();
	else
		do_save_kill(point, mark);
}

command copy_region() on reg_tab[ALT('w')]
{
	if (_is_rectangle)
		copy_rectangle();
	else {
		prev_cmd = 0;
		do_save_only(point, mark);
		this_cmd = 0;
	}
	highlight_off();
}


/* Delete the current character */
command delete_character() on reg_tab[CTRL('D')], reg_tab[KEYDELETE]
{
	do_delete_characters(1);
}

/* Delete the previous character */
command backward_delete_character() on reg_tab[CTRL('H')], reg_tab[DEL]
{
	if (virtual_space && cursor_to_column > 0)
		virtual_move_back();
	else {
		if (delete_hacking_tabs)
			hack_tabs(-iter);
		do_delete_characters(-1);
	}
}

virtual_move_back()	/* move left in virtual space */
{
	int newcol = cursor_to_column, orig = point;

	if ((curchar() == '\n' || character(point - 1) == '\t') &&
						!(newcol % tab_size))
		newcol -= tab_size;
	else
		newcol--;
	to_virtual_column(newcol);
	if (point < orig)
		point = orig - (curchar() == '\t');
}

do_delete_characters(dir)
{
	int start;

	check_modify(bufnum);
	start = point;
	if (has_arg || prev_cmd == APPENDKILL) {
		point += iter * dir;
		do_save_kill(start, point);
	} else if (!over_mode || dir == 1 || point <= 0)
		delete(start, start + dir);
	else if (character(--point) != '\t' && curchar() != '\n')
		replace(point, ' ');
	iter = 0;
}

command append_next_kill() on reg_tab[ALT(CTRL('W'))]
{
	if (_is_rectangle)
		delete_rectangle();
	else
		this_cmd = APPENDKILL;
}

command rectangle_mode() on cx_tab['#']
{				/* go in or out of rectangle mode */
	_is_rectangle = in_rectangle_mode
			= (has_arg ? (iter != 0) : !in_rectangle_mode);
	iter = 0;
	make_mode();
	if (is_highlight_on())
		highlight_on();
}

command kill_rectangle()
{				/* save and replace w/ spaces */
	pick_rectangle(1, 1, 1);
}

command copy_rectangle()
{
	pick_rectangle(1, 0, 0);
}

command delete_rectangle()
{				/* remove columns */
	pick_rectangle(0, 0, 1);
}

pick_rectangle(copy, clear, remove)	/* copy => push copy of rectangle */
{				/* remove => take out rectangle */
	int oldpoint, t;	/* clear => and put spaces instead */
	short start_col, stop_col;	/* columns to operate on */

	iter = 0;
	if (clear || remove)
		check_modify(bufnum);
	start_col = virtual_column();
	t = point;
	point = mark;
	if (mark == _mark_virtual_pos && _mark_virtual_column >= 0)
		stop_col = _mark_virtual_column;
	else
		stop_col = current_column();
	point = t;
	if (point > mark)
		t = point, point = mark, mark = t;
	if (start_col == stop_col)
		return;
	else if (start_col > stop_col)
		t = start_col, start_col = stop_col, stop_col = t;
	if (copy) {
		check_num_kill_buffers();
		push_kill();
		_kill_width[_num_kill] = stop_col - start_col;
	}
	oldpoint = point;
	for (;;) {			/* for each line */
		force_to_column(start_col);
		t = point;
		force_to_column(stop_col);
		if (copy)
			xfer(_cur_kill_buf, t, point);
		if (remove)
			delete(t, point);
		if (clear)
			insert_to_column(start_col, stop_col);
		if (point >= mark || !nl_forward())
			break;
		if (copy)
			xfer(_cur_kill_buf, point - 1, point);
	}
	mark = point;		/* lower right of rectangle */
	point = oldpoint;	/* upper left */
	_is_rectangle = 1;
	force_to_column(start_col);
}

/* yank the current kill buffer into the current buffer */
command yank() on reg_tab[CTRL('Y')]
{
	retrieve(0);
}

/* pop to the previous kill buffer, and replace region with the new buffer */
command yank_pop() on reg_tab[ALT('y')]
{
	retrieve(1);
}

int _oldfrom = -1, _oldto = -1;	/* region yanked last time */
#define UNDO_FLAG_KILL	1	/* bracket each yank in undo with these */
#define UNDO_FLAG_RETR	2	/* for yank-pop */

retrieve(withpop)	/* insert text from kill buffer */
{
	char *thisbuf = bufname;

	check_modify(bufnum);
	check_num_kill_buffers();
	if (_kill_width[_num_kill] == UNUSED)	/* no kills yet */
		return;
	if (withpop && !pop_retrieve())
		error("Must follow yank or yank-pop");
	_oldfrom = point;		/* record start of yank */
	undo_flag = UNDO_FLAG_KILL;	/* mark in undo list */
	bufname = _cur_kill_buf;
	if (bufname == thisbuf)		/* ignore yanking to kill buffer */
		return;
	_is_rectangle = (_kill_width[_num_kill] != REGION);

	if (!_is_rectangle) {
		xfer(thisbuf, 0, size());
		bufname = thisbuf;
		_oldto = point;		/* end of yank */
	} else {			/* yank rectangle */
		rect_retrieve(thisbuf);
		point = _oldfrom;	/* go to start of yank */
		_oldto = mark;		/* end of yank */
	}
	build_first = 1;		/* redisplay hint */
}      

#if 0

pop_retrieve()		/* get previous kill buffer */
{
	int from, to, t, ok = 0, i = 0;
	spot before, after;

	from = point, to = mark;
	if (to < from)		/* ensure from <= to */
		from = mark, to = point;
	undo_flag = UNDO_FLAG_RETR;
	if (from != _oldfrom || to != _oldto)
		return 0;
	do {		/* undo previous yank */
		t = undo_op(1);
		if ((t & UNDO_FLAG) && undo_flag == UNDO_FLAG_KILL) {
			ok = 1 + (_kill_width[_num_kill] == REGION);
			break;		/* found prev yank */
		}
	} while (t && !(t & UNDO_END) && i++ < 5);
	if (ok == 2) {		/* if region */
		before = alloc_spot(0);	/* remember where it was */
		after = alloc_spot(1);
	}
	if (ok != 1) {		/* unless rectangle, redo it now */
		while (!(undo_op(0) & UNDO_FLAG)
					|| undo_flag != UNDO_FLAG_RETR)
			;		/* redo it */
	}
	if (ok == 2) {
		delete(*before, *after);
		free_spot(before);
		free_spot(after);
	} else if (!ok)
		return 0;
	t = _num_kill;
	do {			/* get the previous kill buffer */
		pop_kill();	/* skipping unused kbufs */
	} while (_kill_width[_num_kill] == UNUSED && t != _num_kill);
	return 1;
}

#else

pop_retrieve()		/* get previous kill buffer */
{
	int from, to, t, ok = 0, i = 0;

	from = point, to = mark;
	if (to < from)		/* ensure from <= to */
		from = mark, to = point;
	undo_flag = UNDO_FLAG_RETR;
	if (from != _oldfrom || to != _oldto)
		return 0;
	do {		/* undo previous yank */
		t = undo_op(1);
		if ((t & UNDO_FLAG) && undo_flag == UNDO_FLAG_KILL) {
			ok = 1 + (_kill_width[_num_kill] == REGION);
			break;		/* found prev yank */
		}
	} while (t && !(t & UNDO_END) && i++ < 5);
    if (!ok)
		return 0;
	t = _num_kill;
	do {			/* get the previous kill buffer */
		pop_kill();	/* skipping unused kbufs */
	} while (_kill_width[_num_kill] == UNUSED && t != _num_kill);
	return 1;
}


#endif

rect_retrieve(thisbuf)	/* yank rectangle */
char *thisbuf;
{
	int startcol, stopcol, from, to, t;

	bufname = thisbuf;
	startcol = current_column();
	stopcol = startcol + _kill_width[_num_kill];
	bufname = _cur_kill_buf;
	point = 0;
	while (point < size()) {	/* for each line in kill buf */
		from = point;
		to = nl_forward();
		to = point - to;
		bufname = thisbuf;
		force_to_column(startcol);
		if (over_mode) {	/* delete old stuff */
			t = point;
			force_to_column(stopcol);
			delete(t, point);
		}
		insert_to_column(current_column(), startcol);
		bufname = _cur_kill_buf;
		xfer(thisbuf, from, to);
		bufname = thisbuf;
		if (curchar() != '\n')
			insert_to_column(current_column(), stopcol);
		mark = point;		/* leave mark at end of rect */
		if (!nl_forward())
			insert('\n');
		bufname = _cur_kill_buf;
	}
	bufname = thisbuf;
}

command delete_blank_lines() on cx_tab[CTRL('O')]
{
	int num = 0, c;
	int start;

	check_modify(bufnum);
	re_search(-1, "[ \t\n]*");	/* go to start of this whitespace */
	if (point)			/* skip prev line */
		point += parse_string(1, "[ \t]*\n", NULL);
	start = point;
	while (point < size()) {	/* count lines in whitespace */
		if ((c = curchar()) == '\n')
			num++;
		else if (c != ' ' && c != '\t')
			break;
		point++;
	}
	re_search(-1, "[ \t]*");	/* skip indentation of last line */
	if (num)
		delete(start, point);
	if (!has_arg)			/* with no argument, */
		iter = (num > 1);   /* 1 or 0 blank lines => 0, more => 1 */
	while (iter-- > 0)		/* insert (iter) newlines */
		insert('\n');
	point = start;
}

/* delete spaces and tabs adjacent to point on current line */
command delete_horizontal_space() on reg_tab[ALT('\\')]
{
	check_modify(bufnum);
	re_search(-1, "[ \t]*");		/* to beginning of area */
	delete(point, point + parse_string(1, "[ \t]*", (char *) 0));
}

get_alnum()	/* wait until user presses a letter or digit */
{
	do {
		getkey();
		check_abort();
	} while (!isdigit(key) && !isalpha(key));
	key = toupper(key);
}

command copy_to_scratch() on cx_tab['x']
{
	char buf[80];

	mention("Save region in scratch buffer [A-Z, 0-9]: ");
	get_alnum();
	sprintf(buf, "-temp-buf-%c", key);
	zap(buf);
	xfer(buf, point, mark);
	say("");
	highlight_off();
}

command insert_scratch() on cx_tab['y']
{
	char buf[80];
	int orig = bufnum;

	mention("Restore from scratch buffer [A-Z, 0-9]: ");
	get_alnum();
	sprintf(buf, "-temp-buf-%c", key);
	if (!exist(buf))
		error("Scratch buffer %c is not in use", key);
	mark = point;
	bufname = buf;
	_is_rectangle = 0;
	buf_xfer(orig, 0, size());
	bufnum = orig;
	say("");
}
