/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*	"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.	*
*									*
*  Copyright (C) 1985, 1992 Lugaru Software Ltd.  All rights reserved.	*
*									*
* Limited permission is hereby granted to reproduce and modify this	*
* copyrighted material provided that the resulting code is used only in	*
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.				*
************************************************************************/

#include "eel.h"

buffer char _is_line;

command sort_buffer()
{
	do_sort_buffer(0);
}

command reverse_sort_buffer()
{
	do_sort_buffer(1);
}

do_sort_buffer(rev)	/* ask user where to sort, then do it */
{
	char buf[FNAMELEN];

	save_var sort_status = 1;
	get_buf(buf, "Sort current buffer into buffer", "sorted");
	zap(buf);
	do_buffer_sort(buf, has_arg ? iter : 0, rev);
	to_buffer(buf);
	iter = 0;
}

/* Sort a buffer in place, w/o changing current buffer. */
/* If user aborts & abort_searching is ABORT_ERROR, return ABORT_ERROR */
/* & leave buffer unsorted, else return 0. */

sort_another(buf, col, rev)
char *buf;
{
	int old = bufnum, res;
	char *tmp = temp_buf();

	bufname = buf;
	res = do_buffer_sort(tmp, col, rev);	/* sort into tmp */
	if (!res) {
		zap(buf);	/* replace with (sorted) tmp */
		grab_buffer(name_to_bufnum(tmp));
	}
	bufnum = old;
	delete_buffer(tmp);
	return res;
}

command sort_region()
{
	do_sort_region(point, mark, has_arg ? iter : 0, 0);
	iter = 0;
}

command reverse_sort_region()
{
	do_sort_region(point, mark, has_arg ? iter : 0, 1);
	iter = 0;
}             

#if 0

do_sort_region(from, to, col, rev)	/* sort a region in place */
{
	char *temp1 = temp_buf(), *temp2 = temp_buf(), *orig = bufname;
	char *this_srch_case_map = _srch_case_map;
	int this_case_fold = case_fold, res = ABORT_JUMP;
	jmp_buf this_level;

	save_var sort_status = 1;
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
		xfer(temp1, from, to);
		bufname = temp1;
		case_fold = this_case_fold;
		_srch_case_map = this_srch_case_map;
		res = do_buffer_sort(temp2, col, rev);
		if (!res) {
			bufname = orig;
			delete(from, to);
			bufname = temp2;
			xfer(orig, 0, size());
		}
	}
	bufname = orig;
	delete_buffer(temp1);
	delete_buffer(temp2);
	return res;
}

#else

do_sort_region(from, to, col, rev)	/* sort a region in place */
{
	char *temp1 = temp_buf(), *temp2 = temp_buf(), *orig = bufname;
	char *this_srch_case_map = _srch_case_map;
	int this_case_fold = case_fold, res = ABORT_JUMP, width;
	jmp_buf this_level; 
    save_spot point = to;
    save_spot mark = from;

	save_var sort_status = 1;
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
        if (from > to)
          {
            to = mark;
            from = point;
          }             
        if (_is_rectangle)
          {     
            mark = from;
            point = to;
            width = copy_block(temp1);
          }
        else
          {
            if (_is_line)
              {             
                point = to;
                nl_forward();
                to = point;
                point = from;
                to_begin_line();
                from = point;
              }
            xfer(temp1, from, to);
          }
        bufname = temp1;
		case_fold = this_case_fold;
		_srch_case_map = this_srch_case_map;
		res = do_buffer_sort(temp2, col, rev);
		if (!res) {
			bufname = orig;  
            if (_is_rectangle)
                over_block(width, temp2);
            else
              {
                delete(from, to);
                bufname = temp2;
                xfer(orig, 0, size());
              }
		}
	}
	bufname = orig;
	delete_buffer(temp1);
	delete_buffer(temp2);
	return res;
}

#endif

#if 1

copy_block(tempbuf)
char *tempbuf;
  {
    int   i, j;
    int   width;
    int   StartCol;
    int   EndCol;
    int   SavePoint;
    char  *ThisBuff  = bufname;

    block_setup(&StartCol, &EndCol);
    width = EndCol - StartCol;
    if (width == 0)
        return 0; 
    
    move_to_column(StartCol);
    SavePoint = point;
    
    while (point <= mark)
      {
        i = point;
        move_to_column(EndCol);
        j = point - i;
        
        xfer(tempbuf, i, point);
        bufname = tempbuf;
        i = point;
        point = size();
        if (character(point) == '\n')
            replace(point, ' ');
        if (j != width)
          {
            while (j++ < width)
                insert(' ');
            point = size();
          }
        insert('\n');
        bufname = ThisBuff;
        
        nl_forward();
        move_to_column(StartCol);
      }
    mark = point;
    point = SavePoint; 
    return width;
  }

over_block(width, tempbuf)
char *tempbuf;
  {
    int   i, end, column;
    char  *ThisBuff = bufname;

    column = current_column();
    bufname = tempbuf;
    point = 0;
    end = size();

    while (point < end)
      {   
        bufname = ThisBuff;
        if (current_column() < column)
            to_column(column);
        else
          {
            i = point;
            move_to_column(column + width);
            delete(i, point);
          }
        bufname = tempbuf;
        i = point;
        point += width;
        xfer(ThisBuff, i, point);
        point++;
        bufname = ThisBuff;
        if (!nl_forward())
            insert('\n');
        move_to_column(column);
        bufname = tempbuf;
      }
    bufname = ThisBuff;
  }   
  

#endif

/* Sort the buffer into buf on column col. */
/* If rev is nonzero, do a reverse sort. */
/* If abort_searching is ABORT_ERROR and user aborts, returns ABORT_ERROR. */
/* Otherwise returns 0. */

do_buffer_sort(buf, col, rev)
char *buf;
{
	char rev_map[256];
	int i;

	if (sort_case_fold == 0 || sort_case_fold == 1)
		save_var case_fold = sort_case_fold;
	if (rev) {
		for (i = 0; i < 256; i++)	/* set up for reverse sort */
			rev_map[i] = 255 - _srch_case_map[i];
		save_var _srch_case_map = rev_map;
		save_var case_fold = 1;
	}
	return buffer_sort(buf, col);
}
